/* <hide>
  CCF: a Conditional Compilation Facility
  Copyright 1996-2000 Ian Wild

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  Bug reports, bug fixes, enhancement request, and arbitrarily
  large amounts of chocolate may be sent to the author at <imw@acm.org>.
</hide> */
/* <doc>
<html><head><title>CCF's Symbol Table</title></head>
<body><h1>CCF's Symbol Table</h1>

<p>Currently I keep all the defined names in one doubly-linked list.  If this
proves to be a bottleneck it can be replaced with something more fancy
without upsetting the rest of the program.

</doc> */

#include	<ctype.h>
#include	<stdio.h>
#include	<string.h>

#include	"iofns.h"
#include	"memfns.h"
#include	"syms.h"


/* symbol table entry */

typedef struct Symbol
{
  char	*name;
  long	 value;

  struct Symbol	*next;
  struct Symbol	*prev;
} Symbol;


static	Symbol	*base, *last;

static Symbol *find_entry (char *name)
{
  Symbol *p;
  for (p = base; p; p = p -> next)
    if (strcmp (p -> name, name) == 0)
      break;
  return (p);
}


int IsDefined (char *name)
{
  return (!! find_entry (name));
}

long GetValue (char *name)
{
  Symbol *p = find_entry (name);
  if (! p)
    FatalPrintf (1, "uninitialised variable (%.100s) used", name);
  return (p -> value);
}

void Unset (char *name)
{
  Symbol *p = find_entry (name);
  if (p)
  {
    my_free (p -> name);
    p -> name = NULL;
    if (p -> next)
      p -> next -> prev = p -> prev;
    else
      last = p -> prev;
    if (p -> prev)
      p -> prev -> next = p -> next;
    else
      base = p -> next;
  }
}

void SetValue (char *name, long value)
{
  Symbol *p = find_entry (name);
  if (! p)
  {
    p = my_malloc (sizeof (Symbol));
    p -> name = my_malloc (1 + strlen (name));
    strcpy (p -> name, name);
    p -> next = base;
    if (base)
      base -> prev = p;
    p -> prev = NULL;
    base = p;
  }
  p -> value = value;
}

void ListSymbols (void)
{
  Symbol *p;
  for (p = base; p; p = p -> next)
    fprintf (stderr, "%s: %ld\n", p -> name, p -> value);
}
